import React, { useState } from "react";
import s from "./UsagePanel.module.css";
import c from "classnames";

type UsagePanelProps = {
  fields: {
    [key: string]: string;
  };
  usageMap: {
    [key: string]: Array<HTMLElement>;
  };
  proxyUsageList: {
    [key: string]: boolean;
  };
};

/*
 * Renders a UsagePanel element which shows each field & how it is used on the page
 * @public
 */
const UsagePanel = (props: UsagePanelProps) => {
  const { fields } = props;
  const [isVisible, setVisible] = useState(false);
  const [sortedColIdx, setSortedColIdx] = useState(1);
  const [sortDirection, setSortDirection] = useState("Ascending");

  // Clicking the "sort" button for the same `sortedColIdx` will reverse the sort.
  // Clicking the "sort" button for a different `sortedColIdx` will sort by that column instead.
  const toggleSortState = (colIdx: number) => {
    if (colIdx === sortedColIdx) {
      setSortDirection(
        sortDirection === "Ascending" ? "Descending" : "Ascending"
      );
    } else {
      setSortDirection("Ascending");
      setSortedColIdx(colIdx);
    }
  };

  // aggregate data from field data + `usageMap` + `proxyUsageList` to `UsagePanelFieldProps`
  const getFieldProps = (field: [string, string]): UsagePanelFieldProps => {
    const [fieldKey, fieldVal] = field;
    let shortKey = fieldKey;
    const prefix = "document.streamOutput.";
    if (shortKey.startsWith(prefix)) {
      shortKey = shortKey.slice(prefix.length);
    }

    let shortVal = fieldVal;
    const maxStrLen = 40;
    if (typeof shortVal === "string" && shortVal.length > maxStrLen) {
      shortVal = `${shortVal.slice(0, maxStrLen)}...`;
    }

    const fieldIsUsed = !!(
      props.usageMap[fieldKey] || props.proxyUsageList[fieldKey]
    );
    let fieldStatus = "NotUsed";
    if (props.usageMap[fieldKey]) {
      fieldStatus = "Used";
    } else if (props.proxyUsageList[fieldKey]) {
      fieldStatus = "Internal";
    }

    const usageEls = props.usageMap[fieldKey];

    return {
      fieldKey: shortKey,
      fieldVal: shortVal.toString(),
      fieldIsUsed,
      fieldStatus,
      usedEls: usageEls,
    };
  };

  const getSortFnFromState = () => {
    const colIdxToPropKey: { [key: number]: string } = {
      0: "fieldStatus",
      1: "fieldKey",
    };

    const fieldStatusSortIndices: { [key: string]: string } = {
      Used: "0",
      Internal: "1",
      NotUsed: "2",
    };

    return (a: UsagePanelFieldProps, b: UsagePanelFieldProps) => {
      let a_value =
        a[colIdxToPropKey[sortedColIdx] as keyof UsagePanelFieldProps];
      let b_value =
        b[colIdxToPropKey[sortedColIdx] as keyof UsagePanelFieldProps];

      if (!(a_value && b_value)) {
        return 0;
      }

      // Field statuses should not be sorted alphabetically, instead use the pre-defined order above
      if (colIdxToPropKey[sortedColIdx] === "fieldStatus") {
        a_value = fieldStatusSortIndices[a_value as string];
        b_value = fieldStatusSortIndices[b_value as string];
      }

      if (sortDirection === "Ascending") {
        return a_value > b_value ? 1 : -1;
      } else {
        return a_value < b_value ? 1 : -1;
      }
    };
  };

  return (
    <div className={c(s.panel, { [s.isVisible]: isVisible })}>
      <div className={s.headingRow}>
        <div className={s.heading}>
          Status
          <button className={s.sortButton} onClick={() => toggleSortState(0)}>
            <span
              className={c(s.sortIcon, s[`is${sortDirection}`], {
                [s.isSorted]: sortedColIdx === 0,
              })}
            ></span>
          </button>
        </div>
        <div className={s.heading}>
          Field Data
          <button className={s.sortButton} onClick={() => toggleSortState(1)}>
            <span
              className={c(s.sortIcon, s[`is${sortDirection}`], {
                [s.isSorted]: sortedColIdx === 1,
              })}
            ></span>
          </button>
        </div>
      </div>
      <div className={s.scrollable}>
        {Object.entries(fields)
          .map((field) => getFieldProps(field))
          .sort(getSortFnFromState())
          .map((fieldProps, idx) => (
            <UsagePanelField key={idx} {...fieldProps} />
          ))}
      </div>
      <button
        className={s.toggleBtn}
        onClick={() => setVisible(!isVisible)}
      ></button>
    </div>
  );
};

type UsagePanelFieldProps = {
  fieldKey: string;
  fieldVal: string;
  fieldIsUsed: boolean;
  fieldStatus: string;
  usedEls?: Array<HTMLElement>;
};

const UsagePanelField = (props: UsagePanelFieldProps) => {
  const { fieldKey, fieldVal, fieldIsUsed, fieldStatus, usedEls } = props;

  const statusToDisplayedStatus: { [key: string]: string } = {
    NotUsed: "Not Used",
    Used: "On Page",
    Internal: "Internal Logic",
  };

  // Clicking a field should reveal its usage location on the page
  const showUsageLocation = () => {
    if (usedEls && usedEls.length > 0) {
      const el = usedEls[0];
      document
        .querySelectorAll(`.${s.highlightedUsage}`)
        .forEach((el) => el.classList.remove(s.highlightedUsage));
      el.classList.add(s.highlightedUsage);
      window.scrollTo({
        top: el.getBoundingClientRect().top,
        behavior: "smooth",
      });
    }
  };

  c(s.fieldStatusIndicator, `is${fieldStatus}`);
  return (
    <div
      className={c(
        s.field,
        { [s.isUsed]: fieldIsUsed },
        { [s.isNotUsed]: !fieldIsUsed }
      )}
      onClick={() => showUsageLocation()}
    >
      <span className={c(s.fieldStatusIndicator, s[`is${fieldStatus}`])}></span>
      <span className={s.fieldStatus}>
        {statusToDisplayedStatus[fieldStatus]}
      </span>
      <span className={s.fieldSummary}>
        <span className={s.fieldKey}>{fieldKey}</span>
        <span className={s.fieldVal}>{fieldVal}</span>
      </span>
    </div>
  );
};

export default UsagePanel;
