import { makeNumber, randomize, flattenObj, getDOMNodeAddress } from "./utils";
import ReactDOM from "react-dom";
import React from "react";
import UsagePanel from "./UsagePanel";
import s from "./UsageTooltip.module.css";

type Usage = {
  selector?: string;
  key?: string;
  usageType?: string;
};

/*
 * Set up a MutationObserver which will re-run the CFDebugger code if the DOM changes.
 *  This will catch any React re-renders and make sure the newly rendered elements
 *  have the appropriate markup.
 * @param rerender: the React render function for the root React element
 * @param watchElement: the root element in the DOM to watch
 */
type rerenderFn = (props: Record<any, any>, container?: HTMLElement) => void;
function enableWatchDOMForChanges(
  rerender: rerenderFn,
  watchElement: Node,
  rssProps: Record<any, any>
) {
  const observerConfig = { attributes: true, childList: true, subtree: true };
  const observer = new MutationObserver((mutationsList, observer) => {
    mutationsList.forEach((mutation) => {
      // Don't infinite loop if detecting attribute mutation & attribute is a CFD-attribute
      if (mutation.attributeName && mutation.attributeName.includes("cfd")) {
        return;
      }

      // No new nodes added, so we don't need to update anything
      if (mutation.addedNodes.length < 1) {
        return;
      }

      // Temporarily disable the observer to prevent infinite recursion with react re-renders
      observer.disconnect();

      // Remove any data-attributes or elements added by previous runs of the CFDebugger
      disableAndCleanupCFDebugger();

      // Re-render the React App with randomized props. This has to be done in the real
      //  DOM to accurately track React State changes which have happened since page load.
      // If we try to clone the DOM & render in virtual DOM, the State is reset to default values.
      //  This breaks in the case where the State is causing DOM elements to be hidden/shown, like
      //  for example the FAQ component expand/collapse.
      const [fakeProps, flatProps] = generateRandomizedProps(rssProps);
      rerender(fakeProps); // <- this visibly updates the page - it is the only way to preserve the existing React state.
      const usageList = findFieldValueUsages(flatProps, document.body);

      // Fix the visible page by re-rendering with the real data
      rerender(rssProps);

      // And re-enable all of the CFDebugger features
      addDataAttributeTags(usageList, document.body);
      enableTooltipVisualization();
      enableUsageVisualization(rssProps);

      // Re-enable the observer to watch for any future changes
      observer.observe(watchElement, observerConfig);
    });
  });

  observer.observe(watchElement, observerConfig);
}

/*
 * Remove all created data attributes & DOM Nodes
 */
function disableAndCleanupCFDebugger() {
  document.querySelectorAll("[data-cfd-tooltip]").forEach((el) => {
    if (!(el instanceof HTMLElement)) {
      return;
    }

    Object.entries(el.dataset).forEach(([key, _]) => {
      if (
        key.includes("cfdSource") ||
        key.includes("cfdUsageType") ||
        key.includes("cfdTooltip")
      ) {
        delete el.dataset[key];
      }
    });
  });

  document.querySelectorAll(".CFD-tooltip").forEach((el) => el.remove());
  document.querySelectorAll(".CFD-usagePanel").forEach((el) => el.remove());
}

/*
 * Returns Array of two items:
 *  1. `rssProps` with random values for all fields
 *  2. Flattened list of all randomized Props
 */
function generateRandomizedProps(rssProps: Record<any, any>) {
  const randomizedProps = randomize(rssProps);
  return [randomizedProps, flattenObj(randomizedProps)];
}

/*
 * Returns a list of `usage` objects which describe where and how a `fieldValue` is used
 */
function findFieldValueUsages(
  fieldValues: Record<any, any>,
  sourceDOM: HTMLElement
) {
  const usageList: Array<Usage> = [];

  interface NodeWithUsage extends Node {
    cfdUsage?: string;
  }

  Object.entries(fieldValues).forEach(([key, fieldValue]) => {
    // Create a filtered Node Iterator which will find locations in `sourceDOM` where `fieldValue` occurs
    const nodeIterator = document.createNodeIterator(
      sourceDOM,
      NodeFilter.SHOW_ALL,
      (node) => {
        // The fieldValue could be textContent for a node
        if (
          node.nodeType === Node.TEXT_NODE &&
          node.textContent &&
          node.textContent.includes(fieldValue)
        ) {
          (node as NodeWithUsage).cfdUsage = "Text Content";
          return NodeFilter.FILTER_ACCEPT;

          // The fieldValue could be an attribute on a node
        } else if (node.nodeType === Node.ELEMENT_NODE) {
          if (!(node instanceof HTMLElement)) {
            return NodeFilter.FILTER_REJECT;
          }

          const attrs = node.getAttributeNames();
          for (const attrName of attrs) {
            if (node.getAttribute(attrName) === fieldValue) {
              (node as NodeWithUsage).cfdUsage = `Attribute: ${attrName}`;
              return NodeFilter.FILTER_ACCEPT;
            }
          }
        }

        // We didn't find the fieldValue in this node
        return NodeFilter.FILTER_REJECT;
      }
    );

    // Use the iterator to construct `usageList`
    let nodeWithData;
    while ((nodeWithData = nodeIterator.nextNode())) {
      // Build a selector for elements in `sourceDOM` using :nth-child selectors
      let addressSelector = getDOMNodeAddress(nodeWithData);
      usageList.push({
        selector: addressSelector,
        key: key,
        usageType: (nodeWithData as NodeWithUsage).cfdUsage,
      });
    }
  });

  return usageList;
}

/*
 * Using the param `usageList`, modify the `targetDOM` to add data-attributes
 */
function addDataAttributeTags(usageList: Array<Usage>, targetDOM: HTMLElement) {
  usageList.forEach((usageInfo) => {
    if (!usageInfo.selector) {
      return;
    }
    let el = targetDOM.querySelector(usageInfo.selector);

    if (!(el instanceof HTMLElement)) {
      return;
    }
    el.dataset.cfdTooltip = "true";
    // We need to add a random suffix here so that if a div uses two different
    //  fields, there won't be namespace conflicts in the data-attributes
    const randKey = makeNumber(12);
    el.dataset[`cfdSource_${randKey}`] = usageInfo.key;
    el.dataset[`cfdUsageType_${randKey}`] = usageInfo.usageType;
  });
}

/*
 * Initializes the Custom Field Debugger by detecting how/where each field is used
 *  on the page and marking up the HTML with data-attributes. Renders "Hover tooltip"
 *  and "Panel" UI elements using the data-attributes.
 * @param renderFunction - ReactDOM.render() for the page
 * @param rssProps - props for renderFunction
 * @public
 */
function initCFDebugger(
  renderFunction: rerenderFn,
  rssProps: Record<any, any>
) {
  // Render the initial state of the page with random props in a Virtual DOM
  const [fakeProps, flatFakeProps] = generateRandomizedProps(rssProps);
  const VDOM = document.cloneNode(true) as Document;
  const vRoot = VDOM.querySelector("#root");
  if (!(vRoot instanceof HTMLElement)) {
    return;
  }
  renderFunction(fakeProps, vRoot);

  // Find and tag any field usages
  const usageList = findFieldValueUsages(flatFakeProps, VDOM.body);
  addDataAttributeTags(usageList, document.body);

  // Render the UI
  enableTooltipVisualization();
  enableUsageVisualization(rssProps);

  // Watch for updates
  const documentRoot = document.querySelector("#root");
  if (documentRoot) {
    enableWatchDOMForChanges(renderFunction, documentRoot, rssProps);
  }
}

/*
 * Renders a hover tooltip UI element for each field usage on the page
 */
function enableTooltipVisualization() {
  document.querySelectorAll("[data-cfd-tooltip]").forEach((el) => {
    if (!(el instanceof HTMLElement)) {
      return;
    }
    // Parse the data stored in the data-attributes cfdSource and cfdUsageType
    let cfdInfo: {
      [key: string]: {
        source?: string;
        usage?: string;
      };
    } = {};
    Object.entries(el.dataset).forEach(([key, val]) => {
      if (key.includes("cfdSource")) {
        const [_, id] = key.split("_");
        if (!cfdInfo[id]) {
          cfdInfo[id] = {};
        }
        cfdInfo[id].source = el.dataset[key];
      } else if (key.includes("cfdUsageType")) {
        const [_, id] = key.split("_");
        if (!cfdInfo[id]) {
          cfdInfo[id] = {};
        }
        cfdInfo[id].usage = el.dataset[key];
      }
    });

    // Create & render a new div when the element is hovered
    const toolTipEl = document.createElement("div");
    toolTipEl.classList.add(s.tooltip);
    let innerHTML = "";
    Object.entries(cfdInfo).forEach(([key, val]) => {
      let shortSource = val.source;
      const prefix = "document.streamOutput.";
      if (shortSource?.startsWith(prefix)) {
        shortSource = shortSource.slice(prefix.length);
      }
      innerHTML =
        innerHTML +
        `<div class="${s.section}">
        <span class="${s.source}">Data Field: ${shortSource}</span>
        <span class="${s.usage}">Usage: ${val.usage}</span>
        </div>`;
    });
    toolTipEl.innerHTML = innerHTML;

    el.addEventListener("mouseenter", () => {
      const elRect = el.getBoundingClientRect();
      toolTipEl.style.top = `${elRect.bottom + window.scrollY}px`;
      toolTipEl.style.left = `${elRect.left}px`;
      document.body.appendChild(toolTipEl);
    });
    el.addEventListener("mouseleave", () => {
      toolTipEl.remove();
    });
  });
}

/*
 * Renders a Panel UI element which lists all fields & how/if they are used on the page
 */
function enableUsageVisualization(rssProps: Record<any, any>) {
  const flatProfile = flattenObj(rssProps);

  const cfEls = document.querySelectorAll("[data-cfd-tooltip]");
  const usageMap: Record<string, Array<HTMLElement>> = {};

  Object.entries(flatProfile).forEach(([profileKey, _]) => {
    // Search the page to see if this custom field is in use by checking the 'data-cfd-source' attribute
    cfEls.forEach((el) => {
      if (!(el instanceof HTMLElement)) {
        return;
      }

      Object.entries(el.dataset).forEach(([_, tooltipVal]) => {
        if (profileKey === tooltipVal) {
          usageMap[profileKey] = [
            ...(usageMap[profileKey] ? usageMap[profileKey] : []),
            el,
          ];
        }
      });
    });
  });

  // Add a container element to the DOM to render a react element into
  let usagePanelContainerEl = document.createElement("div");
  usagePanelContainerEl.id = "usagePanelContainer";
  document.body.appendChild(usagePanelContainerEl);

  ReactDOM.render(
    React.createElement(UsagePanel, {
      fields: flatProfile,
      usageMap: usageMap,
      proxyUsageList: window.cfdProxyUsage,
    }),
    usagePanelContainerEl
  );
}

export { initCFDebugger };
